@@ -6,7 +6,7 @@ use base 'XML::SAX::Base';
 use Math::Matrix;
 use Math::Trig qw/deg2rad/;
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 our $GroupId = 'SVGTransformer';
 
 my $IdMatrix = Math::Matrix->id(4);
@@ -24,37 +24,16 @@ sub start_element {
     my $self = shift;
     my $elem = $_[0];
     my $name = lc $elem->{LocalName};
-    if ($name eq 'svg' && !$self->_seen($name)) {
-        $self->_stash(svg => $elem);
+    if ($name eq 'svg' && !$self->_seen($name) && !$self->_stash($name)) {
+        $self->_stash(svg    => $elem);
+        $self->_stash(prefix => $elem->{Prefix});
         return;
-    } elsif ($self->_stash('svg') && !$self->_stash('grouped')) {
-        my $svg = $self->_stash('svg');
-        $self->_stash(grouped => 1);
-        $self->_stash(svg     => undef);
+    } elsif ($self->_stash('svg') && !$self->_stash('updated')) {
+        my @args;
         if ($name eq 'g' && (_attr($elem, 'id') || '') eq $self->_group_id) {
-            $self->_update_tags($svg, $elem);
-            return;
-        } else {
-            my $name = $svg->{Prefix} ? "$svg->{Prefix}:g" : "g";
-            my $group = {
-                LocalName    => 'g',
-                Name         => $name,
-                Prefix       => $svg->{Prefix},
-                NamespaceURI => $svg->{NamespaceURI},
-                Attributes   => {
-                    '{}id' => {
-                        Name  => 'id',
-                        Value => $self->_group_id,
-                    },
-                    '{}transform' => {
-                        Name  => 'transform',
-                        Value => '',
-                    },
-                },
-            };
-            $self->_stash(added_group => 1);
-            $self->_update_tags($svg, $group);
+            push @args, $elem;
         }
+        $self->_update_tags(@args);
     }
     $self->_push($name);
     $self->SUPER::start_element(@_);
@@ -74,42 +53,45 @@ sub comment {
 }
 
 sub end_element {
-    my $self = shift;
-    my $elem = $_[0];
-    my $name = lc $elem->{LocalName};
-    my $prev = $self->{_stack}[-1] || '';
-    if ($name eq 'svg') {
-        my $left = 1;
-        $left++ if $self->_stash('added_wrapper');
-        if ($self->_seen($name) == $left) {
-            my $group_name = $elem->{Prefix} ? "$elem->{Prefix}:g" : 'g';
-            my $group      = {
-                LocalName    => 'g',
-                Name         => $group_name,
-                Prefix       => $elem->{Prefix},
-                NamespaceURI => $elem->{NamespaceURI},
-            };
-
-            if ($self->_stash('added_group') && $prev eq 'g') {
-                $self->_pop($prev);
-                $self->SUPER::end_element($group);
-            }
-            if ($self->_stash('added_wrapper')) {
-                $self->_pop($name);
-                $self->SUPER::end_element(@_);
-                $self->_pop($prev);
-                $self->SUPER::end_element($group);
-            }
-        }
+    my $self   = shift;
+    my $elem   = $_[0];
+    my $name   = lc $elem->{LocalName};
+    my $popped = $self->_pop($name);
+    while (ref $popped) {
+        $self->_end_added_elements($popped);
+        $popped = $self->_pop($name);
     }
-    $self->_pop($name);
     $self->SUPER::end_element(@_);
 }
 
+sub end_document {
+    my $self    = shift;
+    my @stacked = @{$self->{_stack}};
+    return unless @stacked;
+    while (my $popped = pop @stacked) {
+        die "Broken! got: $popped left:" . join(",", @stacked) unless ref $popped;
+        $self->_end_added_elements($popped);
+    }
+}
+
+sub _end_added_elements {
+    my ($self, $tags) = @_;
+    for my $tag (reverse @$tags) {
+        my $prefix = $self->_stash('prefix');
+        $self->SUPER::end_element({
+            LocalName => $tag,
+            Name      => $prefix ? "$prefix:$tag" : $tag,
+            Prefix    => $prefix,
+        });
+    }
+}
+
 sub info { shift->{_info} || {} }
 
 sub _update_tags {
-    my ($self, $svg, $group) = @_;
+    my ($self, $group) = @_;
+
+    my $svg = $self->_delete_stash('svg');
 
     my $svg_width   = _attr($svg, 'width');
     my $svg_height  = _attr($svg, 'height');
@@ -120,24 +102,47 @@ sub _update_tags {
         $svg_viewbox = "0 0 $svg_width $svg_height";
     }
 
-    my $transform = _attr($group, 'transform');
-
     my $view;
-    @$view{qw/min_x min_y max_x max_y tx ty/} = (0, 0, 0, 0, 0, 0);
+    @$view{qw/min_x min_y max_x max_y tx ty w h/} = (0) x 8;
 
     $svg_viewbox = $self->{_comment} if $self->{_comment};
 
     if ($svg_viewbox) {
-        @$view{qw/min_x min_y max_x max_y/} = _split($svg_viewbox);
+        @$view{qw/min_x min_y w h/} = _split($svg_viewbox);
+        $view->{max_x}              = $view->{min_x} + $view->{w};
+        $view->{max_y}              = $view->{min_y} + $view->{h};
     } else {
-        $view->{max_x} = _numify($svg_width);
-        $view->{max_y} = _numify($svg_height);
+        $view->{max_x} = $view->{w} = _numify($svg_width);
+        $view->{max_y} = $view->{h} = _numify($svg_height);
+    }
+
+    my $new_group;
+    if (!$group) {
+        my $prefix = $svg->{Prefix} ? "$svg->{Prefix}:" : "";
+        $group = {
+            LocalName    => 'g',
+            Name         => $prefix . 'g',
+            Prefix       => $svg->{Prefix},
+            NamespaceURI => $svg->{NamespaceURI},
+            Attributes   => {
+                '{}id' => {
+                    Name  => 'id',
+                    Value => $self->_group_id,
+                },
+                '{}transform' => {
+                    Name  => 'transform',
+                    Value => '',
+                },
+            },
+        };
+        $new_group = 1;
     }
-    _translate($view);
 
     my $scale = $self->_scale($view, @$self{qw/Width Height/});
     push @{$self->{_ops}}, ['scale', @$scale];
 
+    my $transform = _attr($group, 'transform');
+
     $self->_parse_transform($transform);
 
     my $matrix = $self->_parse_transform($self->{Transform});
@@ -147,18 +152,18 @@ sub _update_tags {
 
     push @{$self->{_ops}}, ['translate', @$view{qw/tx ty/}];
 
-    my $width  = $view->{max_x};
-    my $height = $view->{max_y};
+    my $width  = $view->{w};
+    my $height = $view->{h};
 
     if ($self->{KeepAspectRatio}) {
         $width  = $self->{Width};
         $height = $self->{Height};
         my @offset = (0, 0);
-        if ($width > $view->{max_x}) {
-            $offset[0] = ($width - $view->{max_x}) / 2;
+        if ($width > $view->{w}) {
+            $offset[0] = ($width - $view->{w}) / 2;
         }
-        if ($height > $view->{max_y}) {
-            $offset[1] = ($height - $view->{max_y}) / 2;
+        if ($height > $view->{h}) {
+            $offset[1] = ($height - $view->{h}) / 2;
         }
 
         if ($offset[0] or $offset[1]) {
@@ -182,33 +187,29 @@ sub _update_tags {
 
             _attr($group, 'id', undef);
 
-            $self->_push('svg');
-            $self->_push('g');
+            $self->_push(['svg', 'g']);
             $self->SUPER::start_element($wrapper);
             $self->SUPER::start_element($wrapper_group);
-            $self->_stash(added_wrapper => 1);
         }
     }
 
-    _attr($svg, 'width',   $view->{max_x});
-    _attr($svg, 'height',  $view->{max_y});
-    _attr($svg, 'viewBox', "0 0 $view->{max_x} $view->{max_y}");
+    _attr($svg, 'width',   $view->{w});
+    _attr($svg, 'height',  $view->{h});
+    _attr($svg, 'viewBox', "0 0 $view->{w} $view->{h}");
 
     $self->_push('svg');
     $self->SUPER::start_element($svg);
     $self->SUPER::comment({Data => $svg_viewbox});
 
     $transform = $self->_ops_to_transform;
-    if ($transform) {
-        _attr($group, 'transform', $transform);
-        $self->_push('g');
+    _attr($group, 'transform', $transform);
+    if ($new_group && $transform) {
+        $self->_push(['g']);
         $self->SUPER::start_element($group);
-    } else {
-        if ($self->_stash('added_group')) {
-            $self->_stash(added_group => undef);
-        }
     }
 
+    $self->_stash(updated => 1);
+
     $self->{_info} = {
         width   => $width,
         height  => $height,
@@ -341,8 +342,8 @@ sub _flatten {
     my @values = map { _numify($_) } (
         $array->[0][0],
         $array->[1][0],
-        $array->[0][1],
-        $array->[1][1],
+        $array->[0][1] - $array->[0][0],
+        $array->[1][1] - $array->[1][0],
         $array->[0][2] + $array->[0][3],
         $array->[1][2] + $array->[1][3],
     );
@@ -352,12 +353,12 @@ sub _flatten {
 sub _translate {
     my $set = shift;
 
-    if ($set->{min_x} && $set->{min_x} < 0) {
+    if ($set->{min_x}) {
         $set->{max_x} -= $set->{min_x};
         $set->{tx}    -= $set->{min_x};
         $set->{min_x} = 0;
     }
-    if ($set->{min_y} && $set->{min_y} < 0) {
+    if ($set->{min_y}) {
         $set->{max_y} -= $set->{min_y};
         $set->{ty}    -= $set->{min_y};
         $set->{min_y} = 0;
@@ -388,6 +389,8 @@ sub _to_hash {
     $hash{max_y} = $y->max->as_array->[0][0];
     $hash{tx}    = $tx->min->as_array->[0][0];
     $hash{ty}    = $ty->min->as_array->[0][0];
+    $hash{w}     = $hash{max_x} - $hash{min_x};
+    $hash{h}     = $hash{max_y} - $hash{min_y};
     \%hash;
 }
 
@@ -396,11 +399,11 @@ sub _scale {
 
     my ($scale_x, $scale_y) = (1, 1);
     if ($x && $y) {
-        if ($set->{max_x}) {
-            $scale_x = $x / $set->{max_x};
+        if ($set->{w}) {
+            $scale_x = $x / $set->{w};
         }
-        if ($set->{max_y}) {
-            $scale_y = $y / $set->{max_y};
+        if ($set->{h}) {
+            $scale_y = $y / $set->{h};
         }
         if ($self->{KeepAspectRatio}) {
             if ($scale_x > $scale_y) {
@@ -410,13 +413,13 @@ sub _scale {
             }
         }
     } elsif ($x) {
-        if ($set->{max_x}) {
-            $scale_x = $x / $set->{max_x};
+        if ($set->{w}) {
+            $scale_x = $x / $set->{w};
         }
         $scale_y = $scale_x;
     } elsif ($y) {
-        if ($set->{max_y}) {
-            $scale_y = $y / $set->{max_y};
+        if ($set->{h}) {
+            $scale_y = $y / $set->{h};
         }
         $scale_x = $scale_y;
     }
@@ -431,7 +434,7 @@ sub _push {
 sub _pop {
     my ($self, $name) = @_;
     my $popped = pop @{$self->{_stack}};
-    if ($name ne $popped) {
+    if (!ref $popped && $name ne $popped) {
         die "Broken! expected: $name got: $popped left:" . join(",", @{$self->{_stack}});
     }
     $popped;
@@ -461,6 +464,11 @@ sub _stash {
     $self->{_stash}{$key};
 }
 
+sub _delete_stash {
+    my ($self, $key) = @_;
+    delete $self->{_stash}{$key};
+}
+
 sub _attr {
     my $elem = shift;
     my $name = shift;
